"""
This class handles the writing of the SQL File
"""


class Fileformatter:

    def __init__(self, file):
        self.file = file

    def write_header(self, database_name):
        """Writes the HEADER(SQL_Mode, time_zone, database name, CREATE and USE)"""
        print("SET SQL_MODE = 'NO_AUTO_VALUE_ON_ZERO';", file=self.file)
        print("SET time_zone = '+00:00';", file=self.file)
        print("", file=self.file)
        print("--", file=self.file)
        print("-- Database: `{}`".format(database_name), file=self.file)
        print("--", file=self.file)
        print(
            "CREATE DATABASE IF NOT EXISTS `{}` DEFAULT CHARACTER SET utf8 COLLATE utf8_general_ci;".format(
                database_name), file=self.file)
        print("USE `{}`;".format(database_name), file=self.file)
        print("", file=self.file)

    def write_table(self, table_name, formatted_columns):
        """Writes table structure"""
        vars_with_no_size = ("double", "datetime", "longtext")
        primary_keys = self._get_primary_keys(formatted_columns)
        index_keys = self._get_index_keys(formatted_columns)

        print("--", file=self.file)
        print("-- Table structure for table `{}`".format(table_name),
              file=self.file)
        print("--", file=self.file)
        print("CREATE TABLE IF NOT EXISTS `{}` (".format(table_name),
              file=self.file)

        counter_column = 0
        for column in formatted_columns:
            column_name, column_var_type, column_var_size, _, column_is_auto_increment = column

            line = "`{}` {}".format(column_name, column_var_type)
            if column_var_type not in vars_with_no_size:
                line += "({})".format(column_var_size)
            line += " NOT NULL"
            if column_is_auto_increment:
                line += " AUTO_INCREMENT"

            if primary_keys.__len__() > 0 or index_keys.__len__() > 0 or counter_column != (
                    formatted_columns.__len__() - 1):
                line += ","
            counter_column += 1
            print(line, file=self.file)

        if primary_keys.__len__() > 0:
            primary_line = "PRIMARY KEY ("
            counter_primary = 0
            for primary_key_name in primary_keys:
                primary_line += "`{}`".format(primary_key_name)
                if counter_primary != (primary_keys.__len__() - 1):
                    primary_line += ", "
                counter_primary += 1
            primary_line += ")"
            if index_keys.__len__() > 0:
                primary_line += ","
            print(primary_line, file=self.file)

        if index_keys.__len__() > 0:
            counter_index = 0
            for index_key_name in index_keys:
                index_line = "KEY `{0}` (`{0}`)".format(index_key_name)
                if counter_index != (index_keys.__len__() - 1):
                    index_line += ","
                counter_index += 1
                print(index_line, file=self.file)

        print(") ENGINE=InnoDB  DEFAULT CHARSET=utf8 AUTO_INCREMENT=1;",
              file=self.file)

    @staticmethod
    def _get_primary_keys(formatted_columns):
        """Get columns that are primary key"""
        primary_keys = list()
        for column in formatted_columns:
            column_is_primary_key = column[3]
            if column_is_primary_key:
                column_name = column[0]
                primary_keys.append(column_name)
        return primary_keys

    @staticmethod
    def _get_index_keys(formatted_columns):
        """Get columns that are index key"""
        index_keys = list()
        for column in formatted_columns:
            column_name = column[0]
            column_is_primary_key = column[3]
            if not column_is_primary_key and column_name.lower().find("id", 0,
                                                                      2) != -1:
                index_keys.append(column_name)
        return index_keys

    def write_table_data(self, table_name, columns, data):
        """Write data of table"""
        print("--", file=self.file)
        print("-- Data for table `{}`".format(table_name), file=self.file)
        print("--", file=self.file)

        for row in data:
            insert_line = "INSERT INTO `{}` (".format(table_name)
            counter_row = 0
            for column in columns:
                column_name = column[3]
                insert_line += column_name
                if counter_row != (columns.__len__() - 1):
                    insert_line += ", "
                counter_row += 1
            insert_line += ") VALUES ("
            counter_value = 0
            for value in row:
                var_type = columns[counter_value][5]
                if isinstance(value, str) or var_type == "DATETIME":
                    insert_line += "'{}'".format(value)
                else:
                    if value is not None:
                        insert_line += "{}".format(value)
                    else:
                        insert_line += "'{}'".format("")
                if counter_value != (columns.__len__() - 1):
                    insert_line += ", "
                counter_value += 1
            insert_line += ");"
            print(insert_line, file=self.file)



"""
This module is used to connect/link to the access file
"""


from os import path
from pyodbc import connect, drivers as pyodrivers, SQL_WVARCHAR, Error


class AccessConnectionError(Exception):
    pass


class ODBCDriverNotFoundError(Exception):
    pass


class Accessconnector:

    @staticmethod
    def driver(file_path):
        """Gets ODBC driver, if not found throw exception"""
        drivers = [x for x in pyodrivers() if x.startswith('Microsoft Access Driver')]
        extension = path.splitext(file_path)[1]
        error_text = """
                     To use this tool you need to have installed 'Microsoft Access Database Engine 2010 Redistributable'.
                     It can be found in the following link: https://www.microsoft.com/en-US/download/details.aspx?id=13255
            
                     'Microsoft Access Database Engine 2010 Redistributable' is a set of components that can be used to facilitate transfer
                     of data between 2010 Microsoft Office System files and non-Microsoft Office applications.
                     """

        if drivers.__len__() == 0:
            raise ODBCDriverNotFoundError(error_text)
        elif extension == ".accdb":
            accdb_drivers = [x for x in drivers if x.endswith('*.accdb)')]
            if accdb_drivers.__len__() == 0:
                raise ODBCDriverNotFoundError(error_text)
            else:
                return accdb_drivers[0]
        else:
            return drivers[0]

    def con(self, driver, file_path, password):
        """Connect to access file, if any error is found during connection, throw exception"""
        try:
            con = connect('DRIVER={};DBQ={};PWD={}'.format(driver, file_path, password))
            con.add_output_converter(SQL_WVARCHAR, self._decode_sketchy_utf16)
            return con
        except Error as ex:
            raise AccessConnectionError(ex.args[1])

    @staticmethod
    def _decode_sketchy_utf16(raw_bytes):
        """Method for ignoring utf-16le chars"""
        s = raw_bytes.decode("utf-16le", "ignore")
        try:
            n = s.index('\u0000')
            s = s[:n]  # respect null terminator
        except ValueError:
            pass
        return s



"""
Class that gets column strucutre, vartypes, etc... intermediary between access file and sql file
"""


from os import path
from tkinter import messagebox
from accessmysqlconverter.fileformatter import Fileformatter


class Accesshandler:
    def __init__(self, cur):
        self._cur = cur

    def make_file(self, output_dir, database_name):
        """Makes sql file"""
        table_names = self._get_table_names()
        suffix = ".sql"
        file = open("{}".format(path.join(output_dir, database_name + suffix)), "w+", encoding='utf-8')
        fileformatter = Fileformatter(file)
        fileformatter.write_header(database_name)
        for name in table_names:
            fileformatter.write_table(name, self._get_columns(name))

        MsgBox = messagebox.askquestion('Import data', 'Do you want to import the data?', icon='warning')
        if MsgBox == "yes":
            for name in table_names:
                fileformatter.write_table_data(name, self._get_table_columns(name), self._get_table_data(name))

        file.close()

    def _get_table_names(self):
        """Retrieves a list with the names of the tables"""
        tables = list()
        for table in self._cur.tables(tableType="TABLE"):
            name = table[2]
            if name[0] != "~":        # Sometimes you get strange tables like ~TR324XZ0 so we need to filter them
                tables.append(name)
        return tables

    def _get_columns(self, table_name):
        """Gets from a table it's columns with it's var names, types..."""
        formatted_columns = list()
        table_columns = self._get_table_columns(table_name)
        table_columns_statistics = self._get_table_columns_statistics(table_name)

        for column in table_columns:
            column_name = column[3]
            column_var_type = column[5]
            column_var_size = column[6]
            column_var_sql_type = self._get_type_SQL(column_var_type)        # Transform ACCESS VARTYPE to SQL VARTYPE
            column_is_primary_key = self._is_primary_key(table_columns_statistics, column_name)
            column_is_auto_increment = column_var_type == "COUNTER"

            formatted_columns.append((column_name, column_var_sql_type, column_var_size, column_is_primary_key, column_is_auto_increment))
        return formatted_columns

    def _get_table_columns(self, table_name):
        """Retrieves a list with the columns of the given table"""
        columns = list()
        cur_columns = self._cur.columns(table=table_name)
        for column in cur_columns:
            columns.append(column)

        return columns

    def _get_table_columns_statistics(self, table_name):
        """Retrieves a list with the columns statistics of the given table (useful for keys)"""
        columns_statistics = list()
        for column in self._cur.statistics(table_name):
            columns_statistics.append(column)
        return columns_statistics

    @staticmethod
    def _get_type_SQL(access_type):
        """Access vartypes to SQL vartypes"""
        switcher = {
            "LONGBINARY": "varbinary",
            "BIT": "bit",
            "COUNTER": "int",
            "CURRENCY": "float",
            "GUID": "int",
            "DOUBLE": "double",
            "INTEGER": "int",
            "BYTE": "int",
            "SMALLINT": "int",
            "DECIMAL": "float",
            "REAL": "int",
            "DATETIME": "datetime",
            "VARCHAR": "varchar",
            "LONGCHAR": "longtext"
        }
        return switcher.get(access_type, "varchar")

    @staticmethod
    def _is_primary_key(table_columns, pColumn_name):
        """Check if column is primary key"""
        for column in table_columns:
            column_name = column[8]
            column_primary_Key = column[5]
            if column_name == pColumn_name and column_primary_Key == "PrimaryKey":
                return True
        return False

    def _get_table_data(self, table_name):
        """Simple query to fetch all data inside a table"""
        print(table_name)
        self._cur.execute("SELECT * FROM `{}`".format(table_name))
        return self._cur.fetchall()



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["pyodbc>=4"]

setup(
    name="accessmysqlconverter",
    version="1.0.4",
    author="miguel93041",
    author_email="miguel.granel.f@gmail.com",
    description="A package to convert MS Access (.mdb, .accdb) into a SQL file (Optionally with It's data)",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/miguel93041/accessmysqlconverter/",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Database :: Database Engines/Servers',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
    include_package_data=True,
    package_data={
        'images': [
            'images/hide_pwd.png',
            'images/watch_pwd.png',
            'images/icon.ico',
        ],
    },
)